#include <stdio.h>
#include <stdlib.h>

/* Binaerer Suchbaum. Implementierung mit */
/* Insert-Funktion ueber Call-by-Value. */

struct tree {
	int value;
	struct tree *left;
	struct tree *right;
};

/* An dieser Stelle ein doppelter Zeiger (struct tree **t), */
/* da der Wert von root (also die Adresse, auf die Root zeigt), */
/* veraendert werden muss. Deshalb wird zunaechst die Adresse von */
/* Root (&root) mit Call-by-Value an die Insert-Funktion uebergeben. */
/* Ueber diesen Adresswert, der nur "lokal" auf dem Stack liegt und nach */
/* beenden der Funktion wieder verloren geht, kann aber der Wert von */
/* root, der zu Beginn auf NULL zeigt, veraendert werden. Wenn also */
/* ein neuer Knoten eingefuegt wird, wird der Wert von root auf einen */ 
/* anderen Speicherbereich zeigen... */
void insert(struct tree * *t, int v) {
	if ((*t) == NULL) {
		(*t) = (struct tree*)calloc(sizeof(struct tree), 1);
		(*t)->value = v;
		(*t)->left = NULL;
		(*t)->right = NULL;
	}
	else if (v <= (*t)->value) {
		insert(&((*t)->left), v);
	}
	else {
		insert(&((*t)->right), v);
	}
}

void print(struct tree *t) {
	if (t != NULL) {
		print(t->left);
		printf("%d\n", t->value);
		print(t->right);
	}
}

int main() {

	struct tree *root = NULL;

	insert(&root, 10);
	insert(&root, 7);
	insert(&root, 1);
	insert(&root, 8);
	insert(&root, 3);
	insert(&root, 5);
	insert(&root, 4);

	print(root);

	return 0;

}
