#include <stdio.h>
#include <stdlib.h>
#include <string.h>

typedef int(*CmpFkt)(const void *, const void *);
typedef void(*PrintFkt)(const void *);
typedef void *(*CopyFkt)(const void *);

/* Damit die Speicherverwaltung fuer die Studentendaten nicht im */
/* Hauptprogramm erfolgt, wird ein neuer Typ CopyFkt definiert, der */
/* Speicherplatz fuer die Studentendaten bereitstellt. Dies ist z.B. */
/* fuer das Loeschen von Knoten sinnvoll, da es sehr aufwendig waere */
/* das Loeschen von Knoten im Hauptprogramm zu implementieren... */

struct studenten {
	int matrnr;
	char name[100];
};

struct tree {
	void *element;
	struct tree *left;
	struct tree *right;
};

void * cpystudis(const void *s) {
    struct studenten *p;

    if ( ((p) = 
		(struct studenten *)calloc(1, sizeof(struct studenten))) == NULL) {
        fprintf(stderr, "error: no memory avaiable!\n");
		exit(1);
    }
    else {
        *p = *((struct studenten *)s);
        return p;
    }
}

int cmpmatrnr(const void *s1, const void *s2) {
    const struct studenten *t1 = s1;
    const struct studenten *t2 = s2;
    return t1->matrnr - t2->matrnr;
}

int cmpname(const void *s1, const void *s2) {
    const struct studenten *t1 = s1;
    const struct studenten *t2 = s2;

    char *n1 = strrchr(t1->name, ' ');
    char *n2 = strrchr(t2->name, ' ');

    int tmp = strcmp(n1, n2);

    if (tmp != 0)
        return tmp;
    return strcmp(t1->name, t2->name);
}

void insert(struct tree * *t, void *s, CopyFkt cpy, CmpFkt cmp) {
	if ((*t) == NULL) {
		if ( ((*t) = (struct tree*)calloc(sizeof(struct tree), 1)) == NULL) {
	        fprintf(stderr, "error: no memory avaiable!\n");
			exit(1);
		}
		else {
			(*t)->element = (*cpy)(s);
			(*t)->left = NULL;
			(*t)->right = NULL;
		}
	}
    else if ( (*cmp)(s, (*t)->element) <= 0 ) {
        insert(&((*t)->left), s, cpy, cmp);
	}
	else {
		insert(&((*t)->right), s, cpy, cmp);
	}
}

void printmatrnr(const void *s) {
    printf("%d\n", ((struct studenten *)s)->matrnr);
}

void printname(const void *s) {
    printf("%s", ((struct studenten *)s)->name);
}

void printstudis(const void *s) {
    printf("%d:%s", ((struct studenten *)s)->matrnr,
        ((struct studenten *)s)->name);
}

void print(struct tree *t, PrintFkt pr) {
	if (t != NULL) {
		print(t->left, pr);
		(*pr)(t->element);
		print(t->right, pr);
	}
}

void discard(struct tree * *t) {
    if ((*t) != NULL) {
        discard(&((*t)->left));
        free((*t)->element);
        (*t)->element = NULL;
        discard(&((*t)->right));
        free(*t);
        (*t) = NULL;
    }
}

int main() {

	struct tree *root = NULL;

	struct studenten studis;

    char *token;
    char buf[30];
    FILE *fp = NULL;

    if (!(fp = fopen("studis.txt", "r"))) {
        fprintf(stderr, "ERROR: FILE OPEN");
        return 1;
    }

    while (fgets(buf, 30, fp) > 0) {
        token = strtok(buf, ":");
        sscanf(token, "%d", &(studis.matrnr));
        token = strtok(NULL, ":");
        strcpy(studis.name, token);
        insert(&root, &studis, cpystudis, cmpmatrnr);
    }
	fclose(fp);

    printf("Nach Matrikelnummer sortiert:\n");

	print(root, printmatrnr);
	print(root, printname);
	print(root, printstudis);

	discard(&root);

	printf("\n");

    if (!(fp = fopen("studis.txt", "r"))) {
        fprintf(stderr, "ERROR: FILE OPEN");
        return 1;
    }

    while (fgets(buf, 30, fp) > 0) {
        token = strtok(buf, ":");
        sscanf(token, "%d", &(studis.matrnr));
        token = strtok(NULL, ":");
        strcpy(studis.name, token);
        insert(&root, &studis, cpystudis, cmpname);
    }
	fclose(fp);

    printf("Nach Matrikelnummer sortiert:\n");

	print(root, printmatrnr);
	print(root, printname);
	print(root, printstudis);

	return 0;

}
